﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using Chapter02_DynamicCalculatorLibrary;
using System.Reflection;

namespace Chapter02
{
    class Program
    {
        static void Main(string[] args)
        {
            CSharp20();
            CSharp30();
            CSharp40();
        }

        #region C# 2.0

        static void CSharp20()
        {
            Generics();
            Nullables();
            Iterators();
        }

        static void Generics()
        {
            List<int> primes = new List<int>();
            for (int n = 2; n <= 100; n++)
            {
                bool isPrime = true;
                for (int d = 2; d <= Math.Sqrt(n); d++)
                {
                    if (n % d == 0)
                    {
                        isPrime = false;
                        break;
                    }
                }

                if (isPrime)
                    primes.Add(n); // statically typed to take in an integer value
            };

            // no performance hit to convert types when retrieving data
            foreach (int prime in primes)
                Console.WriteLine(prime);
        }

        static void Nullables()
        {
            //int a = null; // won’t compile; int is a value type and cannot be set to null
            int? b = null; // valid in C# 2.0 and above

            Console.WriteLine(b ?? 42); // null-coalescing operator (see later)
        }

        static void Iterators()
        {
            foreach (int prime in LazyPrimes())
            {
                Console.WriteLine(prime);

                Console.Write("Calculate next prime? [Y/N] ");
                char key = Console.ReadKey().KeyChar;
                if (char.ToLowerInvariant(key) != 'y')
                    break;

                Console.WriteLine();
            }
        }

        static IEnumerable<int> LazyPrimes()
        {
            for (int n = 2; n <= int.MaxValue; n++)
            {
                bool isPrime = true;
                for (int d = 2; d <= Math.Sqrt(n); d++)
                {
                    if (n % d == 0)
                    {
                        isPrime = false;
                        break;
                    }
                }

                if (isPrime)
                    yield return n; // Yield primes to the caller one-by-one
            };
        }

        #endregion

        #region C# 3.0

        static void CSharp30()
        {
            Linq();
            LocalVariableTypeInference();
            AnonymousTypes();
            InitializerExpressions();
            CollectionInitializers();
            ExtensionMethods();
            LambdaExpressions();
            ExpressionTrees();
            AutoImplementedProperties();
        }

        static void Linq()
        {
            Database db = new Database();

            var res = from product in db.Products
                      where product.Price > 100
                      orderby product.Name
                      select new { product.Name, product.Price };

            foreach (var product in res)
            {
                Console.WriteLine(product);
            }
        }

        class Database
        {
            public Database()
            {
                Products = new List<Product>
                {
                    new Product("Chai") { Price = 123.45m },
                };
            }

            public IEnumerable<Product> Products { get; private set; }
        }

        class Product
        {
            public Product(string name)
            {
                Name = name;
            }

            public string Name { get; set; }
            public decimal Price { get; set; }
        }

        static void LocalVariableTypeInference()
        {
            var whitePages = new Dictionary<string, List<PhoneNumber>>();
            whitePages["Bart"] = new List<PhoneNumber>();
            whitePages["Bart"].Add(new PhoneNumber(1, 425, 0000));
        }

        class PhoneNumber
        {
            public PhoneNumber(int country, int area, int number)
            {
                // Just for illustration purposes...
            }
        }

        static void AnonymousTypes()
        {
            var product = new { Name = "Chai", Price = 123.45m };
            Console.WriteLine(product.Name + " costs " + product.Price);
        }

        static void InitializerExpressions()
        {
            Product p = new Product("Chai") { Price = 123.45m };
            Console.WriteLine(p.Name + " costs " + p.Price);
        }

        static void CollectionInitializers()
        {
            var primes = new List<int> { 2, 3, 5, 7 };
            foreach (var prime in primes)
            {
                Console.WriteLine(prime);
            }
        }

        static void ExtensionMethods()
        {
            string name = "Bart";
            string reverse = Helpers.Reverse(name);
            Console.WriteLine(reverse);
        }

        static class Helpers
        {
            public static string Reverse(string s)
            {
                char[] characters = s.ToCharArray();
                Array.Reverse(characters);
                return new string(characters);
            }
        }

        static void LambdaExpressions()
        {
            // C# 2.0 style
            Func<int, int, int> add20 = delegate(int a, int b) { return a + b; };

            // C# 3.0 style
            Func<int, int, int> add30 = (a, b) => a + b;

            Console.WriteLine(add20(2, 3));
            Console.WriteLine(add30(2, 3));
        }

        static void ExpressionTrees()
        {
            Func<int, int, int> addF = (a, b) => a + b;
            int three = addF(1, 2);
            Console.WriteLine(three);

            Expression<Func<int, int, int>> addE = (a, b) => a + b;
            // the code represented by add can be inspected at runtime
            Console.WriteLine(addE);
        }

        static void AutoImplementedProperties()
        {
            Person bart = new Person("Bart") { Age = 27 };
            Console.WriteLine("{0} is {1}", bart.Name, bart.Age);
        }

        class Person
        {
            public Person(string name)
            {
                Name = name;
            }

            public string Name
            {
                get;
                private set;
            }

            public int Age
            {
                get;
                set;
            }
        }

        #endregion

        #region C# 4.0

        static void CSharp40()
        {
            WithoutDynamic();
            WithDynamic();
            ComInterop();
            CoAndContraVariance();
        }

        static void WithoutDynamic()
        {
            // Ouch, no static type known, so we need to use reflection...
            object calculator = MathTools.GetCalculator();
            int three = (int)calculator.GetType().InvokeMember("Add",
                BindingFlags.InvokeMethod, null, calculator, new object[] { 1, 2 });

            Console.WriteLine(three);
        }

        static void WithDynamic()
        {
            dynamic calculator = MathTools.GetCalculator();
            int three = calculator.Add(1, 2);

            Console.WriteLine(three);
        }

        static void ComInterop()
        {
            // See Chapter 22 for examples
        }

        static void CoAndContraVariance()
        {
            var students = new[]
            {
                new Student("Bart", "CS") { Age = 27 },
                new Student("John", "Languages") { Age = 62 },
            };

            // In C# 4.0, the following compiles due to covariance
            IEnumerable<Person> result = from student in students
                                         where student.Name.StartsWith("B")
                                         select student;

            foreach (var student in result)
            {
                Console.WriteLine(student.Name);
            }

            // In C# 4.0, the following compiles due to contravariance
            IComparer<Person> personComp = new PersonComparerByAge();
            IComparer<Student> studentComp = personComp;

            Console.WriteLine(studentComp.Compare(students[0], students[1]));
        }

        class Student : Person
        {
            public Student(string name, string major)
                : base(name)
            {
                Major = major;
            }

            public string Major { get; private set; }
        }

        class PersonComparerByAge : IComparer<Person>
        {
            public int Compare(Person x, Person y)
            {
                return x.Age - y.Age;
            }
        }


        #endregion
    }
}
