﻿/* Program.cs
*
* Synopsis: Contains the managed entry-point of the application.
* Author: bartde
*
* Change history:
* 06/07/2009 - Initial creation
*/

using System;
using System.Collections;
using System.Collections.Generic;

class Program
{
    static int Main(string[] args)
    {
        foreach (string arg in args)
        {
            Console.WriteLine(arg);
        }

        return args.Length;
    }

    static void PrimerOnTypes()
    {
        var bart = new Person("Bart", 27);
        bart.Print();
    }

    static void Literals()
    {
        // Decimal notation
        {
            byte b = 123;
            short s = -234;
            uint i = 1000000;

            Console.WriteLine("b = " + b);
            Console.WriteLine("s = " + s);
            Console.WriteLine("i = " + i);
        }

        // Hexadecimal notation
        {
            byte b = 0x7B;
            short s = 0xF15;
            uint i = 0x989680;
            Console.WriteLine("b = " + b);
            Console.WriteLine("s = " + s);
            Console.WriteLine("i = " + i);
        }

        // Suffix notation for long (L)
        {
            long res = Math.Abs(-2147483648L);
            Console.WriteLine(res);
        }

        // Character literals
        {
            char c1 = 'A';
            char c2 = 'b';
            char c3 = ' ';
            char c4 = '#';
            Console.WriteLine(c1);
            Console.WriteLine(c2);
            Console.WriteLine(c3);
            Console.WriteLine(c4);

            char c5 = '\t';
            char c6 = '\n';
            char c7 = '\x20'; // a space
            char c8 = '\x2260'; // "not equal to" symbol
            Console.WriteLine(c5);
            Console.WriteLine(c6);
            Console.WriteLine(c7);
            Console.WriteLine(c8);
        }

        // Real
        {
            double half = .5;
            double pi = 3.14159265358979323846; // also available as Math.PI
            Console.WriteLine(half);
            Console.WriteLine(pi);

            double mole = 6.02214179e23;
            double planck = 6.62606896e-34;
            Console.WriteLine(mole);
            Console.WriteLine(planck);
        }

        // Decimal literals with M suffix
        {
            decimal price = 79.95m;
            Console.WriteLine(price);
        }

        // String literals
        {
            string a = "Hello world!";
            string b = "Bart says \"Hello C#!\"";
            string c = @"\\bartde-dev07\c$\Windows\Microsoft.NET\Framework";
            Console.WriteLine(a);
            Console.WriteLine(b);
            Console.WriteLine(c);

            string e = @"<Products>
                             <Product Name=""Chai"" Price=""49.95"" />
                         </Products>";
            Console.WriteLine(e);

            string right = "\x00201";
            Console.WriteLine(right);
        }
    }

    static void Objects()
    {
        try
        {
            ArrayList primes = new ArrayList { 2, 3, 5, 7, 11 };
            int two = (int)primes[0];
            primes.Add("thirtheen");
            int thirteen = (int)primes[5]; // runtime error!
        }
        catch (Exception e)
        {
            Console.WriteLine(e);
        }
    }

    static void Generics()
    {
        List<int> primes = new List<int> { 2, 3, 5, 7, 11 };
        int two = primes[0];
        //primes.Add("thirtheen"); // compile-time error!
        int thirteen = primes[5];
    }

    static void Scope()
    {
        ScopeSample scope = new ScopeSample();
        scope.InnerScope();
        scope.OuterScope();
    }

    public static void PrintAbs(int x)
    {
        if (x >= 0)
        {
            int y = x;
            Print(y);
        }
        else if (x < 0)
        {
            int y = -x;
            Print(y);
        }
    }

    /* Doesn't compile as explained in the code
     * 
    public static void PrintAbs(int x)
    {
        int y = x;
        if (x >= 0)
        {
            Print(y);
        }
        else if (x < 0)
        {
            // Error: A local variable named ‘y’ cannot be declared in this
            // scope because it would give a different meaning to ‘y’,
            // which is already used in a ‘parent or current’ scope to
            // denote something else.
            int y = -x; // error: hiding y in outer scope
            Print(y);
        }
    }
     * 
     */

    /* Doesn't compile as explained in the code
     * 
    private static void PrintInt(int x)
    {
        if (x >= 0)
        {
            int y = x;
            Print(y);
            return;
        }
        // Error: A local variable named ‘y’ cannot be declared in this scope
        // because it would give a different meaning to ‘y’, which is
        // already used in a ‘child’ scope to denote something else.
        int y = -x;
        Print(y);
    }
     * 
     */

    static void Print(int y)
    {
        Console.WriteLine(y);
    }

    static void ImplicitTyping()
    {
        var name = "Bart"; // Tip: hover over var.
        var lotteryNumbers = new Dictionary<string, List<int>>();
        var person = new { Name = "Bart", Age = 26 };

        Console.WriteLine(name);
        Console.WriteLine(lotteryNumbers.Count);
        Console.WriteLine(person);
    }

    static void Preprocessor(string[] args)
    {
        // Switch the editor to Release mode (see toolbar) and observe the following
        // code to become grayed out.
#if DEBUG
        Console.WriteLine("Some diagnostic output:");
        Console.WriteLine(DateTime.Now);
#endif
        foreach (string arg in args)
        {
            Console.WriteLine(arg);
        }
    }

    static void Arrays()
    {
        // Assign and retrieve
        {
            int[] squares = new int[10];
            squares[3] = 9;
            Console.WriteLine(squares[3]); // prints 9

            for (int i = 0; i < squares.Length; i++)
            {
                squares[i] = i * i;
            }
        }

        // Initializers
        {
            int[] squares1 = new int[] { 0, 1, 4, 9, 16, 25, 36, 49, 64, 81 };
            int[] squares2 = new[] { 0, 1, 4, 9, 16, 25, 36, 49, 64, 81 };
            int[] squares3 = { 0, 1, 4, 9, 16, 25, 36, 49, 64, 81 };
            var squares4 = new[] { 0, 1, 4, 9, 16, 25, 36, 49, 64, 81 };
        }

        // Jagged
        {
            int[][] vectors = new int[3][];
            vectors[0] = new int[4];
            vectors[1] = new int[2];
            vectors[2] = new int[3];
        
            var jagged = new[] {
                new [] {
                    new [] { 1, 2, 3 },
                    new [] { 4, 5 }
                },
                new [] {
                    new [] { 6 },
                    new [] { 7, 8, 9 }
                },
            };
        }

        // Multidimensional
        {
            int[,] boardGame = {
                { 0, 1, 2, 3 },
                { 1, 2, 3, 4 },
                { 2, 3, 0, 1 },
                { 3, 0, 1, 2 },
            };
        }
    }

    static void Null()
    {
        string name = null;
        Console.WriteLine(name ?? "unknown"); // Null-coalescing operator
    }

    static void Nullable()
    {
        // Generic type notation
        {
            int a = 42;
            //int b = null; // doesn’t compile
            Nullable<int> c = 42;
            Nullable<int> d = null; // this is valid
            Console.WriteLine(a);
            Console.WriteLine(c);
            Console.WriteLine(d);
        }

        // Kleene notation
        {
            int? c = 42;
            int? d = null; // this is valid
            Console.WriteLine(c);
            Console.WriteLine(d);

            double? number = GetNumber();
            if (number != null)
            {
                // Do something
                Console.WriteLine(number);
            }
        }
    }

    static double? GetNumber()
    {
        return 12.34;
    }
}

class Person
{
    private string _name;
    private int _age;

    public Person(string name, int age)
    {
        _name = name;
        _age = age;
    }

    public void Print()
    {
        Console.WriteLine(_name + " is " + _age + " years young");
    }
}

public class Numbers
{
    // Comment out the attribute to see the CLS compliance warning
    // (see AssemblyInfo.cs for the attribute being declared for the project).
    [CLSCompliant(false)]
    public ulong Large { get; set; }
}

class ScopeSample
{
    private int x = 42;

    public void InnerScope()
    {
        int x = 53;
        int y = x; // refers to the local variable x
        Console.WriteLine(y);
    }

    public void OuterScope()
    {
        int y = x; // refers to the field x
        Console.WriteLine(y);
    }
}

/// <summary>
/// Provides simple calculator functionality.
/// </summary>
public static class Calculator
{
    /// <summary>
    /// Adds the integer numbers <paramref name="a"/>
    /// and <paramref name="b"/> together.
    /// </summary>
    /// <param name="a">First integer term to add.</param>
    /// <param name="b">Second integer term to add.</param>
    /// <returns>
    /// Sum of the integer numbers <paramref name="a"/>
    /// and <paramref name="b"/>
    /// </returns>
    /// <exception cref="System.OverflowException">
    /// Thrown when the computed sum does not fit in the precision of
    /// System.Int32.
    /// </exception>
    /// <example>
    /// <code>int res = Add(1, 2); // produces value 3</code>
    /// </example>
    public static int Add(int a, int b)
    {
        return a + b;
    }
}