﻿using System;
using System.ComponentModel;
using System.Diagnostics;
using System.Runtime.InteropServices;

namespace Chapter12
{
    class Program
    {
        static void Main(string[] args)
        {
            // Call a sample method here.
        }

        static void Constructors_Parameters()
        {
            var bart = new Person("Bart", 27);
        }

        static void Constructors_Default()
        {
            var bart = new Person2();
            bart.Name = "Bart";
            bart.Age = 27;
        }

        static void Constructors_Factory()
        {
            var ticket = Ticket.PurchaseTicket();
        }

        static void Constructors_Structs()
        {
            var pt = new Point(2, 3);

            Console.WriteLine(pt.X);
            Console.WriteLine(pt.Y);
            Console.WriteLine(pt.Z); // 0
        }

        static void Constructors_Static()
        {
            A a = new A();
            Console.WriteLine(a.ToString());

            // Here goes the static constructor...
            A.B();
        }

        static void Finalizers()
        {
            int gen0 = 0;
            while (true)
            {
                B b = new B();

                var new0 = GC.CollectionCount(0);
                if (gen0 != new0)
                {
                    Console.WriteLine("Gen0 collection triggered");
                    gen0 = new0;
                }
            }
        }

        static void Jobs()
        {
            using (var job = Job.CreateOrOpen("Demo"))
            {
                for (int i = 0; i < 5; i++)
                {
                    using (var proc = Process.Start("notepad.exe"))
                        job.AddProcess(proc);
                }

                Console.WriteLine("Press ENTER to terminate job...");
                Console.ReadLine();
                job.Terminate(0);
            }
        }
    }

    class Person
    {
        private readonly string _name;
        private readonly int _age;

        public Person(string name, int age)
        {
            _name = name;
            _age = age;
        }
    }

    class Person2
    {
        private string _name;
        private int _age;

        public string Name
        {
            get { return _name; }
            set { _name = value; }
        }

        public int Age
        {
            get { return _age; }
            set { _age = value; }
        }
    }

    class Ticket
    {
        private readonly int _id;

        private Ticket(int id)
        {
            _id = id;
        }

        public static Ticket PurchaseTicket()
        {
            // A long-running task may be needed to purchase a ticket, ultimately
            // returning one that can only be constructed here:
            return new Ticket(321);
        }
    }

    struct Point
    {
        private readonly int _x, _y, _z;
        private readonly DateTime _t;

        public Point(int x, int y, int z, DateTime t)
        {
            _x = x; _y = y; _z = z; _t = t;
        }

        public Point(int x, int y, int z)
            : this(x, y, z, DateTime.Now)
        {
        }

        public Point(int x, int y)
            : this(x, y, 0, DateTime.Now)
        {
        }

        public int X { get { return _x; } }
        public int Y { get { return _y; } }
        public int Z { get { return _z; } }
    }

    struct A
    {
        static A()
        {
            Console.WriteLine("In A::.cctor");
        }

        public static void B() { }
    }

    class B
    {
        private static int s_id;
        private int _id;

        public B()
        {
            _id = s_id++;
            Console.WriteLine("Allocated instance " + _id);
        }

        ~B()
        {
            Console.WriteLine("Finalized instance " + _id);
        }
    }

    public class Job : IDisposable
    {
        [DllImport("kernel32.dll", CharSet = CharSet.Unicode, SetLastError = true)]
        static extern IntPtr CreateJobObject(SECURITY_ATTRIBUTES att, string name);

        [DllImport("kernel32.dll", SetLastError = true)]
        static extern bool AssignProcessToJobObject(IntPtr job, IntPtr process);

        [DllImport("kernel32.dll", SetLastError = true)]
        static extern bool TerminateJobObject(IntPtr job, uint exitCode);

        [DllImport("kernel32.dll", SetLastError = true)]
        static extern bool CloseHandle(IntPtr obj);

        private bool _disposed;

        private IntPtr _handle;

        private Job(IntPtr handle)
        {
            _handle = handle;
        }

        public static Job CreateOrOpen(string name)
        {
            if (name == null)
                throw new ArgumentNullException("name");

            var res = CreateJobObject(null, name);
            if (res == IntPtr.Zero)
                throw new Win32Exception(); // Will use the last Win32 error.

            return new Job(res);
        }

        public void AddProcess(Process process)
        {
            if (process == null)
                throw new ArgumentNullException("process");

            if (_disposed)
                throw new ObjectDisposedException("job");

            if (!AssignProcessToJobObject(_handle, process.Handle))
                throw new Win32Exception(); // Will use the last Win32 error.
        }

        public void Terminate(uint exitCode)
        {
            if (_disposed)
                throw new ObjectDisposedException("job");

            if (!TerminateJobObject(_handle, exitCode))
                throw new Win32Exception(); // Will use the last Win32 error.
        }

        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        protected virtual void Dispose(bool disposing)
        {
            if (!_disposed)
            {
                if (disposing)
                {
                    // We were called by the Dispose method.
                    // Release managed resources (like other IDisposables) here.
                }

                // Also null out reference type fields here.
                CloseHandle(_handle);
                _disposed = true;
            }

            // If there were a base class that’s IDisposable, call that one too.
            // base.Dispose(disposing);
        }

        ~Job()
        {
            Dispose(false);
        }
    }

    [StructLayout(LayoutKind.Sequential)]
    internal class SECURITY_ATTRIBUTES
    {
        public int nLength;
        public IntPtr lpSecurityDescriptor;
        public int bInheritHandle;
    }
}
