﻿using System;
using System.ComponentModel;
using System.Globalization;

namespace Chapter13
{
    class Program
    {
        static void Main(string[] args)
        {
            // Call a sample method here.
        }

        static void Vector()
        {
            Vector v1 = new Vector(1, 2, 3);
            Vector v2 = new Vector(3, 2, 1);

            Vector allFour = v1 + v2; // (4, 4, 4)
            Console.WriteLine(allFour);

            Vector larger = v1 * 4;   // (4, 8, 12)
            Console.WriteLine(larger);

            int dotProduct = v1 * v2; // 10
            Console.WriteLine(dotProduct);
        }

        static void LiftedNull_BuiltIn()
        {
            int? a = 3, b = 5, c = null;

            // Will print 8. The type of a + b is by itself an int?.
            Console.WriteLine(a + b);

            // All of those will simply print an empty line since the results are null.
            Console.WriteLine(a + c);
            Console.WriteLine(c + b);
            Console.WriteLine(c + c);
        }

        static void LiftedNull_Custom()
        {
            Vector2? v1 = new Vector2(3, 4);
            Vector2? v2 = new Vector2(2, 1);
            Vector2? v3 = null;

            // Will print (5, 5). The type of v1 + v2 is by itself an Vector?.
            Console.WriteLine(v1 + v2);

            // All of those will simply print an empty line since the results are null.
            Console.WriteLine(v1 + v3);
            Console.WriteLine(v3 + v2);
            Console.WriteLine(v3 + v3);
        }

        static void Increment()
        {
            var c = new Cursor(42);
            Console.WriteLine("Before: " + c);
            Console.WriteLine("During: " + c++);
            Console.WriteLine("After: " + c);

            var d = new Cursor(42);
            Console.WriteLine("Before: " + d);
            Console.WriteLine("During: " + ++d);
            Console.WriteLine("After: " + d);
        }

        static void Compound()
        {
            Vector2 v1 = new Vector2(3, 4);
            Vector2 v2 = new Vector2(2, 1);

            // Turns into v1 = v1 + v2;
            v1 += v2;

            Console.WriteLine(v1);
        }

        static void ShortCircuit()
        {
            bool and = PromptBool("l: ") & PromptBool("r: ");
            Console.WriteLine("& -> " + and);
            bool andAlso = PromptBool("l: ") && PromptBool("r: ");
            Console.WriteLine("&& -> " + andAlso);
            bool or = PromptBool("l: ") | PromptBool("r: ");
            Console.WriteLine("| -> " + or);
            bool orElse = PromptBool("l: ") || PromptBool("r: ");
            Console.WriteLine("|| -> " + orElse);
            int res = PromptBool("condition: ") ? Get("true") : Get("false");
            Console.WriteLine("?: -> " + res);
        }

        static bool PromptBool(string s)
        {
            Console.Write(s);
            return bool.Parse(Console.ReadLine());
        }

        static int Get(string s)
        {
            Console.WriteLine(s);
            return 42;
        }

        static void Equality()
        {
            var p1 = new Point(1, 2);
            var p2 = new Point(2, 3);
            var p3 = new Point(1, 2);
            var s = "Other";

            // False
            Console.WriteLine(p1.Equals(p2));
            Console.WriteLine(p2.Equals(p1));
            Console.WriteLine(p1 == p2);
            Console.WriteLine(p2 == p1);
            Console.WriteLine(p1.Equals(s));
            Console.WriteLine(s.Equals(p1));
            Console.WriteLine(p1.Equals(null));
            // Won't compile, null doesn't have a type
            // Console.WriteLine(null.Equals(p1));

            // Shouldn’t compile as there’s no == operator between Point and string
            // Console.WriteLine(p1 == s);
            // Console.WriteLine(s == p1);

            // True
            Console.WriteLine(p1.Equals(p3));
            Console.WriteLine(p3.Equals(p1));
            Console.WriteLine(p1 == p3);
            Console.WriteLine(p3 == p1);
        }

        static void Conversions()
        {
            var pt = new CartesianPoint(1, 1);
            Console.WriteLine(pt);
            Console.WriteLine((PolarPoint)pt);
            Console.WriteLine((CartesianPoint)(PolarPoint)pt);
        }

        static void ConvertClass()
        {
            int five = Convert.ToInt32(5); // Essentially a no-op.
            int nine = Convert.ToInt32("9"); // Much like int.Parse.
            int four = Convert.ToInt32((byte)4); // Much like an implicit cast.
            int zero = Convert.ToInt32(false); // Cannot be achieved using a cast.
            int oops = Convert.ToInt32(DateTime.Now); // InvalidCastException.
        }

        static void Convertible()
        {
            var pt = new Point(3, 4);
            var s = Convert.ToString(pt, null /* format provider */);
        }
    }

    class Vector
    {
        private int _x, _y, _z;

        public Vector(int x, int y, int z)
        {
            _x = x;
            _y = y;
            _z = z;
        }

        // Whitespace is allowed between "operator" and "-".
        public static Vector operator -(Vector vector)
        {
            return new Vector(-vector._x, -vector._y, -vector._z);
        }

        public static Vector operator +(Vector left, Vector right)
        {
            return new Vector(left._x + right._x, left._y + right._y, left._z + right._z);
        }

        // We choose the dot product to show a different return type.
        public static int operator *(Vector left, Vector right)
        {
            return left._x * right._x + left._y * right._y + left._z * right._z;
        }

        // For a scalar product, just one of the operands has to be a Vector.
        public static Vector operator *(Vector vector, int scale)
        {
            return new Vector(vector._x * scale, vector._y * scale, vector._z * scale);
        }

        public override string ToString()
        {
            return string.Format("({0}, {1}, {2})", _x, _y, _z);
        }
    }

    struct Vector2
    {
        private int _x, _y;

        public Vector2(int x, int y)
        {
            _x = x; _y = y;
        }

        public static Vector2 operator +(Vector2 left, Vector2 right)
        {
            return new Vector2(left._x + right._x, left._y + right._y);
        }

        public override string ToString()
        {
            return string.Format("(X = {0}, Y = {1})", _x, _y);
        }
    }

    class Cursor
    {
        private int _pos;

        public Cursor(int pos)
        {
            _pos = pos;
        }

        public static Cursor operator ++(Cursor c)
        {
            // The definition of Cursor is left as an exercise. A simple implementation
            // could just encapsulate an integer value. More advanced examples of a
            // cursor include a "caret" rendered in some custom textbox control, where
            // increment means moving the "caret" to the right.
            // Just return a new cursor that has advanced one position.
            // Contrast to C++’s overloading for this operator, one isn’t asked to
            // change the incoming Cursor’s position in place.
            return new Cursor(c._pos + 1);
        }
    }

    class Person
    {
        public int Id { get; set; }
        public string Name { get; set; }
        public int Age { get; set; }

        public override bool Equals(object obj)
        {
            if (obj == null || obj.GetType() != this.GetType())
                return false;

            var other = (Person)obj;
            return this.Id == other.Id && this.Name.Equals(other.Name);
        }

        public override int GetHashCode()
        {
            // Assume each Person has a unique Id and Name.
            // The Equals method should use those fields for equality checking.
            return this.Id.GetHashCode() ^ (this.Name != null ? this.Name.GetHashCode() : 0);
        }
    }

    [TypeConverter(typeof(PointConverter))]
    struct Point
    {
        private readonly int _x, _y;

        public Point(int x, int y)
        {
            _x = x; _y = y;
        }

        public override bool Equals(object obj)
        {
            if (obj == null || obj.GetType() != this.GetType())
                return false;
            var other = (Point)obj;
            return this == other;
        }

        public override int GetHashCode()
        {
            return _x.GetHashCode() ^ _y.GetHashCode();
        }

        public static bool operator ==(Point p1, Point p2)
        {
            return p1._x == p2._x && p1._y == p2._y;
        }

        public static bool operator !=(Point p1, Point p2)
        {
            return !(p1 == p2);
        }
    }

    struct CartesianPoint
    {
        private readonly double _x, _y;

        public CartesianPoint(double x, double y)
        {
            _x = x;
            _y = y;
        }

        // Add get-only properties and ToString overload.
        public static explicit operator PolarPoint(CartesianPoint p)
        {
            // Math.Atan could be used below as well, feeding in p._y / p._x. This
            // will still work fine, even if _x is zero, due to the use of floating
            // point PositiveInfinity or NegativeInfinity for the division-by-zero.
            return new PolarPoint(
                Math.Sqrt(p._x * p._x + p._y * p._y),
                Math.Atan2(p._y, p._x)
            );
        }
    }

    struct PolarPoint
    {
        private readonly double _radius, _theta;

        public PolarPoint(double radius, double theta)
        {
            _radius = radius; _theta = theta;
        }

        // Add get-only properties and ToString overload.
        public static explicit operator CartesianPoint(PolarPoint p)
        {
            return new CartesianPoint(
                p._radius * Math.Cos(p._theta),
                p._radius * Math.Sin(p._theta)
            );
        }
    }

    class PointConverter : TypeConverter
    {
        public override bool CanConvertFrom(ITypeDescriptorContext context, Type sourceType)
        {
            if (sourceType == typeof(string))
                return true;
            return base.CanConvertFrom(context, sourceType);
        }

        public override bool CanConvertTo(ITypeDescriptorContext context, Type destinationType)
        {
            if (destinationType == typeof(string))
                return true;
            return base.CanConvertTo(context, destinationType);
        }

        public override object ConvertFrom(ITypeDescriptorContext context, CultureInfo culture, object value)
        {
            if (value == null)
                throw base.GetConvertFromException(value);

            var s = value as string;
            if (s != null)
            {
                int x, y;
                string[] parts = s.Split(',');
                if (parts.Length != 2 || !int.TryParse(parts[0], out x) || !int.TryParse(parts[1], out y))
                    throw base.GetConvertFromException(value);
                return new Point(x, y);
            }
            return base.ConvertFrom(context, culture, value);
        }

        public override object ConvertTo(ITypeDescriptorContext context, CultureInfo culture, object value, Type destinationType)
        {
            if (value is Point && destinationType == typeof(string))
                return ((Point)value).ToString();
            return base.ConvertTo(context, culture, value, destinationType);
        }
    }
}
