﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Linq.Expressions;
using System.Threading;

namespace Chapter17
{
    class Program
    {
        static void Main(string[] args)
        {
            // Call a sample method here.
        }

        static void Basics()
        {
            BinOp add = new BinOp(Add);
            int three = add(1, 2); // calls the Add method through the delegate
        }

        static void MethodGroups()
        {
            BinOp add = Add;
            int three = add(1, 2); // calls the Add method through the delegate
        }

        static int Add(int a, int b)
        {
            return a + b;
        }

        static void InstanceMethod()
        {
            var calc = new Calculator(); // with an Add instance method
            BinOp add = calc.Add;
            int three = add(1, 2); // calls the Add method on calc through the delegate
        }

        static void AnonymousMethod()
        {
            BinOp add = delegate(int a, int b) { return a + b; };
            int three = add(1, 2); // calls the anonymous method through the delegate
        }

        static void Closures()
        {
            int factor = 2;
            BinOp addAndFactor = delegate(int a, int b) { return factor * (a + b); };
            int six = addAndFactor(1, 2);
        }

        static void Closures_LeftValueCapturing()
        {
            int i = 1;
            Action print = delegate() { Console.WriteLine(i); };
            i = 2;
            print(); // print is called after i has been changed to 2
        }

        static void Closures_LoopVariables_Unexpected()
        {
            // Likely (!) prints all 10s
            for (int i = 0; i < 10; i++)
                new Thread(delegate() { Console.WriteLine(i); }).Start();
        }

        static void Closures_LoopVariables_Fix()
        {
            for (int i = 0; i < 10; i++)
            {
                int j = i;
                new Thread(delegate() { Console.WriteLine(j); }).Start();
            }
        }

        static void Lambda()
        {
            BinOp add = (a, b) => a + b;
            int three = add(1, 2); // calls the anonymous method through the delegate
        }

        static void Currying()
        {
            Func<int, Func<int, int>> add = a => b => a + b; // same as a => (b => a + b)
            var three = add(1)(2); // add(1) returns (b => 1 + b) which gets called with 2
            var addFive = add(5); // add(5) returns (b => 5 + b)
            var seven = addFive(2); // addFive(2) returns 5 + 2 = 7
        }

        static void ExpressionTrees()
        {
            Func<int, int, int> addDelegate = (int a, int b) => a + b;
            Console.WriteLine(addDelegate(1, 2)); // Code executable

            Expression<Func<int, int, int>> addExpr = (int a, int b) => a + b;
            Console.WriteLine(addExpr); // Code as data
        }

        static void FuncTypes()
        {
            Func<int, int, int> add = (a, b) => a + b;
            int three = add(arg2: 2, arg1: 2);
        }

        static void ExtensibleCalculator()
        {
            var calc = new Calculator2();

            while (true)
            {
                Console.WriteLine("Operand 1: ");
                int lhs = int.Parse(Console.ReadLine());

                Console.WriteLine("Operator: ");
                string op = Console.ReadLine();

                Console.WriteLine("Operand 2: ");
                int rhs = int.Parse(Console.ReadLine());

                Console.WriteLine("{0}({1},{2}) = {3}", op, lhs, rhs,
                calc.Eval(lhs, op, rhs));
            }
        }

        static void Linq()
        {
            var input = new List<int> { 8, 3, 5, 1, 2, 7, 4 };

            // Imperative
            var result = new List<int>();
            foreach (int x in input)
            {
                if (x % 2 != 0)
                {
                    result.Add(x + 1);
                }
            }

            // Using LINQ
            var res = input.Where(x => x % 2 != 0) // filter on predicate
                           .Select(x => x + 1);    // map using projection

            foreach (var x in res)
                Console.WriteLine(x);
        }

        static void Async_Problem()
        {
            var calc = new Calculator2();

            calc.AddOperator("F1", (a, b) =>
            {
                Thread.Sleep(5000);
                return a + b;
            });

            // Long blocking call!
            calc.Eval(1, "F1", 2); // calls into the custom delegate
        }

        static uint Fibonacci(uint n)
        {
            return n <= 1 ? n : Fibonacci(n - 1) + Fibonacci(n - 2);
        }

        static void Async_Fibonacci_LongRunning()
        {
            for (uint n = 0; n <= 40; n++)
            {
                var watch = new Stopwatch();

                watch.Start();
                uint result = Fibonacci(n);
                watch.Stop();

                Console.WriteLine("Fib({0}) = {1} (in {2} ms)", n, result, watch.ElapsedMilliseconds);
            }
        }

        static void Async_Fibonacci_AsyncInvoke1()
        {
            for (uint n = 1; n <= 20; n++)
            {
                CalculateFib1(n);
            }

            Console.ReadLine();
        }

        static void CalculateFib1(uint n)
        {
            Func<uint> fibN = () => Fibonacci(n);

            var watch = new Stopwatch();
            watch.Start();

            fibN.BeginInvoke(asyncResult =>
            {
                watch.Stop();
                uint result = fibN.EndInvoke(asyncResult);
                Console.WriteLine("Fib({0}) = {1} (in {2} ms)", n, result, watch.ElapsedMilliseconds);
            }, null);

            Console.WriteLine("Computing Fib({0})...", n);
        }

        static void Async_Fibonacci_AsyncInvoke2()
        {
            const int N = 20;

            var waitHandles = new WaitHandle[N];
            for (uint n = 1; n <= N; n++)
            {
                waitHandles[n - 1] = CalculateFib2(n);
            }

            WaitHandle.WaitAll(waitHandles);
        }

        static WaitHandle CalculateFib2(uint n)
        {
            Func<uint> fibN = () => Fibonacci(n);

            var watch = new Stopwatch();
            watch.Start();

            var asyncRes = fibN.BeginInvoke(asyncResult =>
            {
                watch.Stop();
                uint result = fibN.EndInvoke(asyncResult);
                Console.WriteLine("Fib({0}) = {1} (in {2} ms)", n, result, watch.ElapsedMilliseconds);
            }, null);

            Console.WriteLine("Computing Fib({0})...", n);

            return asyncRes.AsyncWaitHandle;
        }

        static void Async_Fibonacci_AsyncInvoke3(uint n)
        {
            Func<uint> fibN = () => Fibonacci(n);

            var asyncResult = fibN.BeginInvoke(_ => { }, null);
            while (!asyncResult.IsCompleted)
            {
                Console.Write(".");
                Thread.Sleep(500);
            }
            Console.WriteLine();

            uint result = fibN.EndInvoke(asyncResult);
            Console.WriteLine("Fib({0}) = {1}", n, result);
        }

        static void Multicast()
        {
            Action<string> sayHiEnglish = name => Console.WriteLine("Hello " + name);
            Action<string> sayHiDutch   = name => Console.WriteLine("Hallo " + name);
            Action<string> sayHiFrench  = name => Console.WriteLine("Allo "  + name);

            Action<string> sendGreetings = null;
            sendGreetings += sayHiEnglish;
            sendGreetings += sayHiDutch;
            sendGreetings("Bart"); // Prints "Hello Bart" and "Hallo Bart"

            sendGreetings -= sayHiEnglish;
            sendGreetings += sayHiFrench;
            sendGreetings("John"); // Prints "Hallo John" and "Allo John"
        }
    }

    delegate int BinOp(int a, int b);

    class Calculator
    {
        public int Add(int a, int b) { return a + b; }
        public int Sub(int a, int b) { return a - b; }
        public int Mul(int a, int b) { return a * b; }
        public int Div(int a, int b) { return a / b; }
        public int Mod(int a, int b) { return a % b; }
    }

    class Calculator2
    {
        private Dictionary<string, Func<int, int, int>> _ops = new Dictionary<string, Func<int, int, int>> {
            { "Add", (a, b) => a + b },
            { "Sub", (a, b) => a - b },
            { "Mul", (a, b) => a * b },
            { "Div", (a, b) => a / b },
            { "Mod", (a, b) => a % b }
        };

        public int Eval(int lhs, string op, int rhs)
        {
            return _ops[op](lhs, rhs);
        }

        public void AddOperator(string name, Func<int, int, int> function)
        {
            _ops[name] = function;
        }
    }
}
