﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Reflection;
using System.Reflection.Emit;
using System.Linq.Expressions;

namespace Chapter21
{
    class Program
    {
        static void Main(string[] args)
        {
            // Call a sample method here.
        }

        static void DynamicInvocation_DynamicTyping()
        {
            dynamic d = GetCalculator();
            int res = d.Add(1, 2);
            Console.WriteLine(res);
        }

        static void DynamicInvocation_Reflection()
        {
            object d = GetCalculator();
            var res = (int)d.GetType().GetMethod("Add").Invoke(d, new object[] { 1, 2 });
            Console.WriteLine(res);
        }

        private static dynamic GetCalculator()
        {
            //
            // Notice we "forget" the static type on the method's return type.
            // This is just a sample; in a realistic scenario, the calculator may
            // come from a world where there simply are no types (e.g. Python).
            //
            return new Calc();
        }

        class Calc
        {
            public int Add(int a, int b)
            {
                return a + b;
            }
        }

        static void TypeOf()
        {
            var stringType = typeof(string);
            var intType = typeof(int);

            Console.WriteLine(stringType.FullName);
            Console.WriteLine(intType.FullName);
        }

        static new void GetType()
        {
            var someObject = "Reflect on me";
            var someType = someObject.GetType();

            Console.WriteLine(someType.FullName);
        }

        static void PrimitiveTypes()
        {
            var res = from t in typeof(string).Assembly.GetTypes()
                      where t.Namespace == "System" && t.IsPrimitive
                      orderby t.Name
                      select t;

            foreach (var t in res)
                Console.WriteLine(t.Name);
        }

        static void GetTypeFromAssembly()
        {
            var int32 = typeof(string).Assembly.GetType("System.Int32");

            Console.WriteLine(int32.FullName);
        }

        static void InstantiateTypes()
        {
            var uri = (Uri)Activator.CreateInstance(typeof(Uri), "http://demo");
            Console.WriteLine(uri);
        }

        static void GenericTypes()
        {
            var intList = typeof(List<>).MakeGenericType(typeof(int));
            var res = (IEnumerable<int>)Activator.CreateInstance(intList, new[] { 1, 3 });

            foreach (var item in res)
                Console.WriteLine(item);
        }

        static void LateBoundMethodInvocation()
        {
            var f = new Foo();

            // Early-bound call
            f.Bar("Early-bound");

            // Late-bound call
            f.GetType().GetMethod("Bar").Invoke(f, new object[] { "Late-bound" });
        }

        class Foo
        {
            public void Bar(string message)
            {
                Console.WriteLine(message);
            }
        }

        static void PropertiesAndIndexers()
        {
            foreach (var p in typeof(string).GetProperties())
            {
                Console.WriteLine(p.Name);

                if (p.CanRead)
                    Console.WriteLine("get = " + p.GetGetMethod());
                if (p.CanWrite)
                    Console.WriteLine("set = " + p.GetSetMethod());
            }
        }

        static void LateBoundPropertyAccess()
        {
            var f = new Foo2();

            // Early-bound set (assume Foo has a Bar property defined)
            f.Bar = 42;

            // Late-bound get
            var res = (int)f.GetType().GetProperty("Bar").GetValue(f, null /* indexer */);
            Console.WriteLine(res);
        }

        class Foo2
        {
            public int Bar { get; set; }
        }

        static void Events()
        {
            var evt = typeof(Foo3).GetEvent("Event");
            Console.WriteLine(evt);
            Console.WriteLine(evt.GetAddMethod());
            Console.WriteLine(evt.GetRemoveMethod());

            var f = new Foo3();
            EventHandler handler = (o, e) => { Console.WriteLine("Raised!"); };

            // Early-bound add
            f.Event += handler;
            f.Raise(); // Prints Raised!

            // Late-bound remove
            evt.RemoveEventHandler(f, handler);
            f.Raise(); // Doesn’t print Raised! anymore
        }

        class Foo3
        {
            public event EventHandler Event;

            public void Raise()
            {
                var e = Event;
                if (e != null)
                    e(this, EventArgs.Empty);
            }
        }

        static void Fields()
        {
            foreach (var f in typeof(DateTime).GetFields())
                Console.WriteLine(f);

            foreach (var f in typeof(DateTime).GetFields(BindingFlags.NonPublic | BindingFlags.Instance))
                Console.WriteLine(f);
        }

        static void Attributes()
        {
            var attrs = typeof(Sample).GetCustomAttributes(typeof(MyAttribute), false);
            var myAttr = (MyAttribute)attrs[0];

            Console.WriteLine(myAttr.Name);
            Console.WriteLine(myAttr.NamedInt);
        }

        [AttributeUsage(AttributeTargets.Class, Inherited = false, AllowMultiple = true)]
        sealed class MyAttribute : Attribute
        {
            public MyAttribute(string name)
            {
                Name = name;
            }

            public string Name { get; private set; }

            public int NamedInt { get; set; }
        }

        [My("Hello", NamedInt = 42)]
        class Sample
        {
        }

        static void LCG_Hello()
        {
            // An assembly...
            var asm = AppDomain.CurrentDomain.DefineDynamicAssembly(
                new AssemblyName("LCGDemo"),
                AssemblyBuilderAccess.Run);

            // ...contains a module...
            var mod = asm.DefineDynamicModule("MyModule");

            // ...contains a type...
            var tpe = mod.DefineType("Sample");

            // ...contains a method...
            var mtd = tpe.DefineMethod("SayHello", MethodAttributes.Public | MethodAttributes.Static);

            // ...contains a code body...
            var gen = mtd.GetILGenerator();
            gen.Emit(OpCodes.Ldstr, "Hello LCG!");

            // The mythical infoof(Console.WriteLine(string))
            gen.Emit(OpCodes.Call, typeof(Console).GetMethod("WriteLine", new[] { typeof(string) }, null));
            gen.Emit(OpCodes.Ret);

            // Makes the code ready to execute...
            var res = tpe.CreateType();
            res.GetMethod("SayHello").Invoke(null /* static */, null /* args */);
        }

        static void LCG_Arith()
        {
            var a = new Param();
            var b = new Param();
            var f = new Function
            {
                Body = new Arithmetic
                {
                    Left = new Arithmetic
                    {
                        Left = a,
                        Right = b,
                        Operation = Operation.Add
                    },
                    Right = new Const
                    {
                        Value = 2
                    },
                    Operation = Operation.Multiply
                },
                Parameters = new List<Param> { a, b }
            };

            var res = f.Compile<Func<int, int, int>>();
            Console.WriteLine(res(2, 3)); // 10 = (2 + 3) * 2
        }

        abstract class Expr
        {
            public abstract void Compile(ILGenerator gen, Dictionary<Param, int> env);
        }

        class Param : Expr
        {
            public override void Compile(ILGenerator gen, Dictionary<Param, int> env)
            {
                gen.Emit(OpCodes.Ldarg, env[this]); // load argument with index...
            }
        }

        class Const : Expr
        {
            public int Value { get; set; }

            public override void Compile(ILGenerator gen, Dictionary<Param, int> env)
            {
                gen.Emit(OpCodes.Ldc_I4, Value);
            }
        }

        class Arithmetic : Expr
        {
            public Expr Left { get; set; }
            public Expr Right { get; set; }
            public Operation Operation { get; set; }

            public override void Compile(ILGenerator gen, Dictionary<Param, int> env)
            {
                Left.Compile(gen, env);
                Right.Compile(gen, env);
                switch (Operation)
                {
                    case Operation.Add:
                        gen.Emit(OpCodes.Add);
                        break;
                    case Operation.Multiply:
                        gen.Emit(OpCodes.Mul);
                        break;
                    case Operation.Subtract:
                        gen.Emit(OpCodes.Sub);
                        break;
                    case Operation.Divide:
                        gen.Emit(OpCodes.Div);
                        break;
                }
            }
        }

        enum Operation
        {
            Add,
            Subtract,
            Multiply,
            Divide
        }

        class Function
        {
            public List<Param> Parameters { get; set; }
            public Expr Body { get; set; }

            public T Compile<T>()
            {
                var asm = AppDomain.CurrentDomain.DefineDynamicAssembly(
                    new AssemblyName("LCG"),
                    AssemblyBuilderAccess.RunAndCollect);

                var mod = asm.DefineDynamicModule("Lightweight");
                var tpe = mod.DefineType("Code");
                var mtd = tpe.DefineMethod("Generation",
                    MethodAttributes.Public | MethodAttributes.Static,
                    // Return type
                    typeof(int),
                    // Every of the Parameters.Count parameters is of type int as well
                    Enumerable.Repeat(typeof(int), this.Parameters.Count).ToArray()
                );

                // Establish environment; we could also use a List<T> and use IndexOf
                // in the recursive compilation code to find the parameter’s index.
                var env = new Dictionary<Param, int>();
                int arg = 0;
                foreach (var par in Parameters)
                    env[par] = arg++;

                // Compilation is requested for the Body, and a final “ret” instruction
                // is added to the IL generator.
                var gen = mtd.GetILGenerator();
                this.Body.Compile(gen, env);
                gen.Emit(OpCodes.Ret);

                var res = tpe.CreateType();

                // Create a delegate of the specified type, referring to the generated
                // method which we always call “Generation” (see the DefineMethod call).
                return (T)(object)Delegate.CreateDelegate(typeof(T), res.GetMethod("Generation"));
            }
        }

        static void ExpressionTrees()
        {
            Expression<Func<int, int, int>> f = (a, b) => (a + b) * 2;

            /*
             * Same as:
             * 
            ParameterExpression a, b;
            Expression<Func<int, int, int>> expression =
                Expression.Lambda<Func<int, int, int>>(
                    Expression.Multiply(
                        Expression.Add(
                            a = Expression.Parameter(typeof(int), "a"),
                            b = Expression.Parameter(typeof(int), "b")
                        ),
                        Expression.Constant(2, typeof(int))
                    ),
                    new[] { a, b }
                );
             */

            Console.WriteLine(f);

            // Compiled at runtime
            var g = f.Compile();
            Console.WriteLine(g(2, 3));

            // Compiled at compile time
            var h = new Func<int, int, int>((a, b) => (a + b) * 2);
            Console.WriteLine(h(2, 3));
        }

        static void ExpressionTrees_Statements()
        {
            var i = Expression.Variable(typeof(int), "i");
            var brk = Expression.Label("Break");
            var a = Expression.Lambda<Action>(
                Expression.Block(
                    new[] { i },
                // int i = 0;
                    Expression.Assign(i, Expression.Constant(0)),
                // while (true)
                    Expression.Loop(
                // {
                        Expression.Block(
                // if
                            Expression.IfThen(
                // (i == 10)
                                Expression.Equal(i, Expression.Constant(10)),
                //     break;
                                Expression.Goto(brk)
                            ),
                // Console.WriteLine(i);
                            Expression.Call(
                                typeof(Console).GetMethod("WriteLine", new[] { typeof(int) }),
                                i
                            ),
                // i++;
                            Expression.PostIncrementAssign(i)
                        ),
                        brk
                // }
                    )
                )
            );

            var f = a.Compile();
            f(); // Prints 0 to 9
        }

        static void ExpressionTrees_Visitors_Find()
        {
            Expression<Func<int, int, int, int>> sum = (x, y, z) => x + y + z + 1;
            var fa = new FindAdditions();
            fa.Visit(sum);
        }

        class FindAdditions : ExpressionVisitor
        {
            protected override Expression VisitBinary(BinaryExpression node)
            {
                if (node.NodeType == ExpressionType.Add)
                    Console.WriteLine(node.Left + " added to " + node.Right);

                return base.VisitBinary(node);
            }
        }

        static void ExpressionTrees_Visitors_Optimizer()
        {
            Expression<Func<bool, bool>> bl = x => !!!!!x;
            Console.WriteLine(bl);
            var blo = new BooleanLogicOptimizer();
            Console.WriteLine(blo.Visit(bl));
        }

        class BooleanLogicOptimizer : ExpressionVisitor
        {
            protected override Expression VisitUnary(UnaryExpression node)
            {
                // If we represent a ! operator...
                if (node.NodeType == ExpressionType.Not)
                {
                    // ...and the recursive visit to our operand...
                    var res = base.Visit(node.Operand);
                    // ...is also a ! operator...
                    if (res.NodeType == ExpressionType.Not)
                    {
                        // ...then we can forget about the double negation:
                        return ((UnaryExpression)res).Operand;
                    }
                }
                return base.VisitUnary(node);
            }
        }
    }
}
