﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Threading;
using System.Diagnostics;

namespace Chapter30
{
    class Program
    {
        static void Main(string[] args)
        {
            // Call a sample method here.
        }

        static void Tasks()
        {
            // Constructor
            {
                var task = new Task<int>(() =>
                {
                    // Do lots of work here, computing int value “res”.
                    var res = 42;
                    return res;
                });

                task.Start();

                // Could do other stuff before blocking for the result.
                Console.WriteLine(task.Result);
            }

            // StartNew
            {
                var task = Task<int>.Factory.StartNew(() =>
                {
                    var res = 42;
                    Thread.Sleep(10000); // Mimick lots of work
                    return res;
                });

                while (!task.Wait(1000)) // Could also use task.IsCompleted
                    Console.Write(".");
                Console.WriteLine(" Done!");

                Console.WriteLine(task.Result);
            }

            // Creation options
            {
                var t = Task.Factory.StartNew(() =>
                {
                    var s = Task.Factory.StartNew(() =>
                    {
                        Console.Write("Task " + Task.CurrentId + " waiting for input... ");
                        return Console.ReadLine();
                    }, TaskCreationOptions.AttachedToParent);

                    // Do work and return.
                });
            }
        }

        static void TaskCancellation()
        {
            var cts = new CancellationTokenSource();

            var task = new Task<int>(() =>
            {
                // Do lots of work here, computing int value “res”.
                var res = 42;

                // Can check the cts.Token whether cancellation has been requested
                // in order to bail out.
                Thread.Sleep(10000); // Mimick lots of work

                return res;
            });

            Console.Write("Waiting for task result...");
            try
            {
                task.Wait(cts.Token);
                Console.WriteLine(task.Result);
            }
            catch (OperationCanceledException)
            {
                Console.WriteLine("Cancelled!");
            }
        }

        static void TaskExceptions()
        {
            var t = Task<int>.Factory.StartNew(() =>
            {
                throw new InvalidOperationException("Oops!");
            });

            try
            {
                var res = t.Result; // Oops, this will throw
            }
            catch (AggregateException ae)
            {
                ae.Flatten().Handle(ex =>
                {
                    var ioe = ex as InvalidOperationException;
                    if (ioe != null)
                    {
                        // Handle the exception here and return true when succeeded.
                        return true;
                    }
                    return false;
                });
            }
        }

        static void TaskContinuations()
        {
            var t = Task<int>.Factory.StartNew(() =>
            {
                throw new InvalidOperationException("Oops!");
            });

            t.ContinueWith(t_ =>
            {
                var ex = t_.Exception;
                // Handle the AggregateException here (using Handle).
            }, TaskContinuationOptions.OnlyOnFaulted);

            // A success continuation (using OnlyOnNotFaulted) could be specified too.
        }

        static void UnhandledExceptions()
        {
            Task.Factory.StartNew(() => { throw new Exception("Oops!"); });

            Console.ReadLine();
            GC.Collect(); // Don’t try this at home!

            /*
             * Instead of GC.Collect(), this would work too, showing the non-deterministic behavior that results.
             * 
            while (true)
            {
                var bs = new byte[1024 * 1024]; // Stressing the GC for our own education.
                Console.WriteLine(i++);
            }
             */
        }

        static void UnobservedExceptions()
        {
            TaskScheduler.UnobservedTaskException += (sender, e) =>
            {
                e.Exception.Flatten().Handle(ex =>
                {
                    // Try to handle the exception, returning true or false.
                    return true;
                });

                // If we get here, we can tell the TPL not to “escalate” on the finalizer.
                e.SetObserved();
            };

            Task.Factory.StartNew(() => { throw new Exception("Oops!"); });

            // Exercise: set a breakpoint in the event handler above.
            Console.ReadLine();
            GC.Collect(); // Don’t try this at home!
        }

        static void ContinuationPassingStyle()
        {
            // Manual plumbing
            Add(2, 3, c =>
            {
                Console.WriteLine(c);
            });

            // Hidden behind BeginInvoke/EndInvoke for delegates
            Func<int, int, int> add = (a, b) => a + b;
            add.BeginInvoke(3, 4, iar =>
            {
                int c = add.EndInvoke(iar);
                Console.WriteLine(c);
            }, /* state */ null);
        }

        static void Add(int a, int b, Action<int> continueWith)
        {
            continueWith(a + b);
        }

        static void ContinuationsOnTasks()
        {
            Task<int>.Factory.StartNew(() => 42)
                .ContinueWith(t => t.Result / 2)
                .ContinueWith(t => t.Result.ToString())
                .ContinueWith(t => Console.WriteLine("42 / 2 = " + t.Result));
        }

        static void Cancellation()
        {
            var cts = new CancellationTokenSource();
            var ct = cts.Token;

            Task.Factory.StartNew(() =>
            {
                while (true)
                {
                    // Alternatively, ThrowIfCancellationRequested could be used,
                    // requiring the caller to handle this case.
                    if (ct.IsCancellationRequested)
                        return;

                    Console.WriteLine(DateTime.Now);
                    Thread.Sleep(1000);
                }
            }, ct);

            Console.WriteLine("Press ENTER to stop the clock...");
            Console.ReadLine();
            cts.Cancel();

            Console.WriteLine("Clock stopped");
            Console.ReadLine();
        }

        static void ParallelInvoke()
        {
            Parallel.Invoke(
                () => Console.WriteLine("First task"),
                () => Console.WriteLine("Second task"),
                () => Console.WriteLine("Third task")
            );
            Console.ReadLine();
        }

        static void WaitAllWaitAny()
        {
            var cts = new CancellationTokenSource();
            ParallelInvoke(cts.Token,
                () => Console.WriteLine("First task"),
                () => Console.WriteLine("Second task"),
                () => Console.WriteLine("Third task")
            );
        }

        static void ParallelInvoke(CancellationToken ct, params Action[] actions)
        {
            var tasks = (from action in actions
                         select Task.Factory.StartNew(action, ct))
                        .ToArray();
            Task.WaitAll(tasks, ct);
        }

        static void ParallelForLoops()
        {
            double[,] xs = GetMatrix(300, 400);
            double[,] ys = GetMatrix(400, 500);
            double[,] zs = new double[300, 500];

            var sw = Stopwatch.StartNew();

            // Parallel
            Parallel.For(0, zs.GetLength(0), i =>
            {
                for (int j = 0; j < zs.GetLength(1); j++)
                {
                    zs[i, j] = 0.0;
                    for (int k = 0; k < xs.GetLength(1); k++)
                    {
                        zs[i, j] += xs[i, k] * ys[k, j];
                    }
                }
            });

            sw.Stop();
            Console.WriteLine(sw.Elapsed);

            sw.Restart();

            // Sequential
            for (int i = 0; i < zs.GetLength(0); i++)
            {
                for (int j = 0; j < zs.GetLength(1); j++)
                {
                    zs[i, j] = 0.0;
                    for (int k = 0; k < xs.GetLength(1); k++)
                    {
                        zs[i, j] += xs[i, k] * ys[k, j];
                    }
                }
            }

            sw.Stop();
            Console.WriteLine(sw.Elapsed);
        }

        static Random s_rand = new Random();

        static double[,] GetMatrix(int a, int b)
        {
            var res = new double[a, b];

            for (int i = 0; i < a; i++)
                for (int j = 0; j < b; j++)
                    res[i, j] = s_rand.Next();

            return res;
        }

        static void ParallelLoops()
        {
            var xs = Enumerable.Range(0, 1000000).ToArray();
            long total = 0;

            Parallel.For<long>(0, xs.Length,
                () => 0,
                (j, loop, subtotal) =>
                {
                    subtotal += xs[j];
                    return subtotal;
                },
                subtotal => Interlocked.Add(ref total, subtotal)
            );

            Console.WriteLine("The total is {0}", total);
        }

        static void ParallelForEach()
        {
            var xs = Enumerable.Range(0, 1000000);
            long total = 0;

            Parallel.ForEach(xs,
                () => 0L,
                (x, loop, subtotal) =>
                {
                    subtotal += x;
                    return subtotal;
                },
                subtotal => Interlocked.Add(ref total, subtotal)
            );

            Console.WriteLine("The total is {0}", total);
        }
    }
}
